#include "main.h"
#include "Convolve.h"
#include "Arithmetic.h"
#include "Geometry.h"

void Convolve::help() {
    printf("\n-convolve takes a frames, width, and height and a 3d matrix specified across\n"
	   "the rows, then down the columns, then over time, and convolves the current\n"
	   "image by that matrix independently in each channel.\n\n"
           "With no arguments, convolve will use the next image on the stack as the filter.\n"
           "It must be single channel.\n\n"
	   "Usage: ImageStack -load a.tga -convolve 1 2 1  -1 1 -save dx.tga\n\n");
	   

}

void Convolve::parse(vector<string> args) {
    if (args.size() > 3) {
        int frames, width, height, size;
        frames = readInt(args[0]);
        width = readInt(args[1]);
        height = readInt(args[2]);
        size = frames * width * height;
        assert((int)args.size() == size + 3, "a size of %ix%ix%i requires %i more arguments. %i were given.", 
               frames, width, height, size, (int)args.size() - 3);
        assert(size % 2 == 1, "filter must have odd size\n");
        
	Image filter(frames, width, height, 1);

	size_t i = 3;
	for (int t = 0; t < frames; t++) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    filter(t, x, y)[0] = readFloat(args[i++]);
		}
	    }
	}
        
        Image im = apply(stack(0), filter);
        pop();
        push(im);
    } else if (args.size() == 0) {
        assert(stack(1).channels == 1, "filter must be single channel\n");
        Image im = apply(stack(0), stack(1));
        pop();
        push(im);
    } else {
        panic("-convolve needs either zero or at least four arguments\n");
    }
}


Image Convolve::apply(Window im, Window filter) {
    int filterSize = filter.frames * filter.width * filter.height;
    assert(filterSize % 2 == 1, "filter must have odd size\n");

    int xoff = (filter.width - 1)/2;
    int yoff = (filter.height - 1)/2;
    int toff = (filter.frames - 1)/2;

    Image out(im.frames, im.width, im.height, im.channels);

    for (int t = 0; t < im.frames; t++) {
        for (int y = 0; y < im.height; y++) {
            for (int x = 0; x < im.width; x++) {
                for (int dt = -toff; dt <= toff; dt++) {
                    int filtert = dt + toff;
                    int imt = MIN(MAX(t + dt, 0), im.frames-1);
                    for (int dy = -yoff; dy <= yoff; dy++) {
                        int filtery = dy + yoff;
                        int imy = MIN(MAX(y + dy, 0), im.height-1);
                        for (int dx = -xoff; dx <= xoff; dx++) {
                            int filterx = dx + xoff;
                            int imx = MIN(MAX(x + dx, 0), im.width-1);                            
                            float weight = filter(filtert, filterx, filtery)[0];
                            for (int c = 0; c < im.channels; c++) {                                
                                out(t, x, y)[c] += weight * im(imt, imx, imy)[c];
                            }
                        }
                    }
                }
            }
        }
    }

    return out;
}

void Deconvolve::help() {
    printf("\n-deconvolve uses a gradient descent to undo a convolution. The convolution\n"
	   "specified must be 3D. The arguments are the amount of time in seconds the the\n"
	   "algorithm should run for, then the filter frames, filter width, filter height,\n"
	   "then the filter to be deconvolved by in row major form. With only the first\n"
	   "argument, deconvolve will use the next image on the stack as the filter. It\n"
	   "must be single channel. The output to deconvolve is constrained to be positive.\n\n"
	   "Usage: ImageStack -load in.jpg -deconvolve 60 3 3 1 1 1 1 1 1 1 1 1 -save sharp.jpg\n\n");
}


void Deconvolve::parse(vector<string> args) {
    if (args.size() > 4) {
        int frames, width, height, size;
	float seconds = readFloat(args[0]);
        frames = readInt(args[1]);
        width = readInt(args[2]);
        height = readInt(args[3]);
        size = frames * width * height;
        assert((int)args.size() == size + 4, "a size of %ix%ix%i requires %i more arguments. %i were given.", 
               frames, width, height, size, (int)args.size() - 4);
        assert(size % 2 == 1, "matrix must have odd size\n");
        
	Image filter(frames, width, height, 1);

	size_t i = 4;
	for (int t = 0; t < frames; t++) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    filter(t, x, y)[0] = readFloat(args[i++]);
		}
	    }
	}
        
	
        Image im = apply(stack(0), filter, seconds);
	pop();
	push(im);
    } else if (args.size() == 1) {
        assert(stack(1).channels == 1, "filter must be single channel\n");
	float seconds = readFloat(args[0]);
        Image im = apply(stack(0), stack(1), seconds);
	pop();
	pop();
	push(im);
    } else {
        panic("-deconvolve needs either one or at least five arguments\n");
    }
}

Image Deconvolve::apply(Window im, Window filter, float seconds) {
    float stepSize = 0.1;


    // start by normalizing the filter
    float filterMagnitude = 0;
    for (int t = 0; t < filter.frames; t++) {
	for (int y = 0; y < filter.height; y++) {
	    for (int x = 0; x < filter.width; x++) {
		filterMagnitude += filter(t, x, y)[0];
	    }
	}
    }

    Scale::apply(filter, 1.0f/filterMagnitude);

    // check the time now
    float finish = currentTime() + seconds;

    Image solution(im);

    // otherwise solve the problem at a coarser scale first to get our estimate
    if (0 && im.width * im.height * im.frames > 100 && filter.width*filter.height*filter.frames > 10) {
	int widthDiv = 2, heightDiv = 2, framesDiv = 2;
	if (im.width == 1 || filter.width == 1) widthDiv = 1;
	if (im.height == 1 || filter.height == 1) heightDiv = 1;
	if (im.frames == 1 || filter.frames == 1) framesDiv = 1;
	Image smallIm = Resample::apply(im, im.frames/framesDiv, im.width/widthDiv, im.height/heightDiv);
	Image smallFilter = Resample::apply(filter, filter.frames/framesDiv, filter.width/widthDiv, filter.height/heightDiv);
	solution = apply(smallIm, smallFilter, seconds/8);
	solution = Resample::apply(solution, im.frames, im.width, im.height);	
    }

    // refine the estimate further by pretending the kernel is separable and deconvolving by that
    // don't do this if the filter is already flat

    // t direction
    if (filter.frames > 1 && (filter.width > 1 || filter.height > 1)) {
	Image flatFilter = Downsample::apply(filter, 1, filter.width, filter.height);
	solution = apply(solution, flatFilter, seconds/8);
    }

    if (filter.height > 1 && (filter.width > 1 || filter.frames > 1)) {
	// y direction
	Image flatFilter = Downsample::apply(filter, filter.frames, filter.width, 1);
	solution = apply(solution, flatFilter, seconds/8);
    }

    if (filter.width > 1 && (filter.frames > 1 || filter.height > 1)) {
	// x direction
	Image flatFilter = Downsample::apply(filter, filter.frames, 1, filter.height);
	solution = apply(solution, flatFilter, seconds/8);
    }

    printf("Deconvolving %ix%ix%i / %ix%ix%i\n", im.frames, im.width, im.height, filter.frames, filter.width, filter.height);

    int xoff = (filter.width - 1)/2;
    int yoff = (filter.height - 1)/2;
    int toff = (filter.frames - 1)/2;

    double error = 1, oldError = 0, oldOldError = 0;
    while (1) {
        oldOldError = oldError;
	oldError = error;
	error = 0;
	for (int t = 0; t < im.frames; t++) {
	    for (int y = 0; y < im.height; y++) {
		for (int x = 0; x < im.width; x++) {
		    for (int c = 0; c < im.channels; c++) {

			float sum = 0;

			// convolve the proposed solution here
			for (int dt = -toff; dt <= toff; dt++) {
			    int filtert = dt + toff;
			    int imt = MIN(MAX(t + dt, 0), im.frames-1);
			    for (int dy = -yoff; dy <= yoff; dy++) {
				int filtery = dy + yoff;
				int imy = MIN(MAX(y + dy, 0), im.height-1);
				for (int dx = -xoff; dx <= xoff; dx++) {
				    int filterx = dx + xoff;
				    int imx = MIN(MAX(x + dx, 0), im.width-1);
				    
				    float weight = filter(filtert, filterx, filtery)[0];
				    sum += weight * solution(imt, imx, imy)[c];
				}
			    }
			}

			// figure out the error
			float delta = im(t, x, y)[c] - sum;
			error += delta * delta;

                        // distribute the error across the filter
			for (int dt = -toff; dt <= toff; dt++) {
			    int filtert = dt + toff;
			    int imt = MIN(MAX(t + dt, 0), im.frames-1);
			    for (int dy = -yoff; dy <= yoff; dy++) {
				int filtery = dy + yoff;
				int imy = MIN(MAX(y + dy, 0), im.height-1);
				for (int dx = -xoff; dx <= xoff; dx++) {
				    int filterx = dx + xoff;
				    int imx = MIN(MAX(x + dx, 0), im.width-1);
				    float weight = filter(filtert, filterx, filtery)[0];       
				    solution(imt, imx, imy)[c] += stepSize * weight * delta;
				    if (solution(imt, imx, imy)[c] < 0) solution(imt, imx, imy)[c] = 0;
				}
			    }
			}

			// continue on to the next pixel
		    } 
		}
	    }
	}


	// do it again in reverse
	for (int t = im.frames-1; t >= 0; t--) {
	    for (int y = im.height-1; y >= 0; y--) {
		for (int x = im.width-1; x >= 0; x--) {
		    for (int c = im.channels-1; c >= 0; c--) {

			float sum = 0;

			// convolve the proposed solution here
			for (int dt = -toff; dt <= toff; dt++) {
			    int filtert = dt + toff;
			    int imt = MIN(MAX(t + dt, 0), im.frames-1);
			    for (int dy = -yoff; dy <= yoff; dy++) {
				int filtery = dy + yoff;
				int imy = MIN(MAX(y + dy, 0), im.height-1);
				for (int dx = -xoff; dx <= xoff; dx++) {
				    int filterx = dx + xoff;
				    int imx = MIN(MAX(x + dx, 0), im.width-1);
				    
				    float weight = filter(filtert, filterx, filtery)[0];
				    sum += weight * solution(imt, imx, imy)[c];
				}
			    }
			}

			// figure out the error
			float delta = im(t, x, y)[c] - sum;
			error += delta * delta;

                        // distribute the error across the filter
			for (int dt = -toff; dt <= toff; dt++) {
			    int filtert = dt + toff;
			    int imt = MIN(MAX(t + dt, 0), im.frames-1);
			    for (int dy = -yoff; dy <= yoff; dy++) {
				int filtery = dy + yoff;
				int imy = MIN(MAX(y + dy, 0), im.height-1);
				for (int dx = -xoff; dx <= xoff; dx++) {
				    int filterx = dx + xoff;
				    int imx = MIN(MAX(x + dx, 0), im.width-1);
				    float weight = filter(filtert, filterx, filtery)[0];       
				    solution(imt, imx, imy)[c] += stepSize * weight * delta;
				    if (solution(imt, imx, imy)[c] < 0) solution(imt, imx, imy)[c] = 0;
				}
			    }
			}

			// continue on to the next pixel
		    } 
		}
	    }
	}


	error /= (im.frames * im.width * im.height * im.channels * 2);
	printf("%f %f\n", error, stepSize); fflush(stdout);
	// some ad hoc momentum to speed things up
        if (error < oldError && oldError < oldOldError) stepSize *= 1.5;
	else if (error <= oldError) stepSize *= 1.2;
	else stepSize *= 0.3;

	if (error < 0.000001 || stepSize < 0.000001) break;

	// check to see if we've run out of time yet	
        if (currentTime() > finish) break;
    }

    return solution;
}



void DeconvolveCG::help() {
}

void DeconvolveCG::parse(vector<string> args) {
    if (args.size() > 4) {
        int frames, width, height, size;
	float seconds = readFloat(args[0]);
        frames = readInt(args[1]);
        width = readInt(args[2]);
        height = readInt(args[3]);
        size = frames * width * height;
        assert((int)args.size() == size + 4, "a size of %ix%ix%i requires %i more arguments. %i were given.", 
               frames, width, height, size, (int)args.size() - 4);
        assert(size % 2 == 1, "matrix must have odd size\n");
        
	Image filter(frames, width, height, 1);

	size_t i = 4;
	for (int t = 0; t < frames; t++) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    filter(t, x, y)[0] = readFloat(args[i++]);
		}
	    }
	}
        
	
        Image im = apply(stack(0), filter, seconds);
	pop();
	push(im);
    } else if (args.size() == 1) {
        assert(stack(1).channels == 1, "filter must be single channel\n");
	float seconds = readFloat(args[0]);
        Image im = apply(stack(0), stack(1), seconds);
	pop();
	pop();
	push(im);
    } else {
        panic("-deconvolvecg needs either one or at least five arguments\n");
    }    
}

Image DeconvolveCG::apply(Window im, Window filter, float seconds) {
    return apply(im, filter, seconds, im);
}

Image DeconvolveCG::apply(Window im, Window filter, float seconds, Window initialGuess) {
    float finish = currentTime() + seconds;

    // TODO: detect if the kernel is separable

    Image target(im);

    // figure out some properties of the kernel
    bool symmetric = true;
    bool positiveDefinite = true;

    for (int t = 0; t < filter.frames && symmetric; t++) {
        for (int y = 0; y < filter.height && symmetric; y++) {
            for (int x = 0; x < (filter.width+1)/2 && symmetric; x++) {
                float val1 = filter(t, x, y)[0];
                float val2 = filter(filter.frames-1-t, filter.width-1-x, filter.height-1-y)[0];
                if (fabs(val1 - val2) > 0.000001) symmetric = false;
            }
        }
    }

    symmetric = true;

    // if the kernel is not symmetric, we need to convolve by it's transpose too
    Image filterTranspose(filter);
    Flip::apply(filterTranspose, 'x');
    Flip::apply(filterTranspose, 'y');
    Flip::apply(filterTranspose, 't');

    if (!symmetric) {
        printf("System is asymmetric, convolving with filter transpose.\n");
        target = Convolve::apply(im, filterTranspose);
    }

    printf("Starting CG for filter size %i %i %i\n", filter.frames, filter.width, filter.height);

    Image x(initialGuess);

    // r = b - Ax
    Image r((Window )target); 
    Image q = Convolve::apply(x, filter); 
    if (!symmetric || !positiveDefinite) q = Convolve::apply(q, filterTranspose);
    Subtract::apply(r, q);

    // d = r
    Image d((Window )r);
    
    double errorNew = norm(r);
    double errorOld = errorNew;

    for (int i = 0; ; i++) {
        // q = Ad
        q = Convolve::apply(d, filter);
        if (!symmetric || !positiveDefinite) q = Convolve::apply(q, filterTranspose);

        float alpha = errorNew / dot(d, q);

        // x = alpha*d + x;
        scaleAdd(x, alpha, d, x);

        if (i % 50 == 49) {
            // remove accumulated error in the residual
            // r = b - Ax
            r = Image(target);
            q = Convolve::apply(x, filter);
            if (!symmetric || !positiveDefinite) q = Convolve::apply(q, filterTranspose);
            Subtract::apply(r, q);
        } else {
            // estimate new residual directly
            // r = -alpha*q + r
            scaleAdd(r, -alpha, q, r);
        }        
        
        errorOld = errorNew;
        errorNew = norm(r);

        // d = beta*d + r
        double beta = errorNew / errorOld;
        scaleAdd(d, beta, d, r);

        printf("Error: %3.10f     \t (%3.1f seconds to go)\n", errorNew/(target.width*target.height*target.frames), finish - currentTime());

        // check if the error is increasing
        if (beta > 1.1) {
            if (positiveDefinite) {
                printf("Error increasing, system must not be positive definite.\n");
                positiveDefinite = false;
                // convolve the target by the filter transpose
                target = Convolve::apply(target, filterTranspose);
            } else {
                printf("Error is increasing for unknown reasons.\n");
            }

            // restart the CG
            x = Image(initialGuess);
            r = Image(target);
            q = Convolve::apply(x, filter);
            if (!symmetric || !positiveDefinite) q = Convolve::apply(q, filterTranspose);
            Subtract::apply(r, q);
            errorNew = norm(r);
            Scale::apply(d, 0);
            Add::apply(d, r);
        }

        if (errorNew < 0.00000000001 || currentTime() > finish) break;
    }    
    
    return x;
}

double DeconvolveCG::dot(Image &a, Image &b) {
    double val = 0;
    float *ptr1 = a(0, 0, 0), *ptr2 = b(0, 0, 0);
    int size = a.frames * a.width * a.height * a.channels;
    for (int i = 0; i < size; i++) {
        val += (*ptr1++) * (*ptr2++);
    }
    return val;
}

double DeconvolveCG::norm(Image &a) {
    double val = 0;
    float *ptr = a(0, 0, 0);
    int size = a.frames * a.width * a.height * a.channels;
    for (int i = 0; i < size; i++) {
        val += (*ptr) * (*ptr);
        ptr++;
    }
    return val;
}

// compute a = alpha*b + c
void DeconvolveCG::scaleAdd(Image &a, float alpha, Image &b, Image &c) {
    float *aPtr = a(0, 0, 0);
    float *bPtr = b(0, 0, 0);
    float *cPtr = c(0, 0, 0);
    int size = a.frames * a.width * a.height * a.channels;
    for (int i = 0; i < size; i++) {
        *aPtr++ = alpha * (*bPtr++) + (*cPtr++);
    }        
}
